﻿using System;
using System.Data;
using System.Configuration;
using System.IO;
using SharpMap.CoordinateSystems;
using SharpMap.CoordinateSystems.Transformations;
using SharpMap.Styles;
using System.Xml;
using System.Drawing;
using System.Drawing.Drawing2D;
using SharpMap.Geometries;
using SharpMap;
using SharpMap.Layers;
using SharpMap.Data.Providers;
using System.Collections.Generic;
using System.Web;

/// <summary>
/// GetMapProcessor 的摘要说明
/// </summary>
public class GetMapProcessor
{
    HttpContext context;
    public GetMapProcessor(HttpContext context)
    {
        this.context = context;
    }

    //部分代码抄袭了SharpMap
    internal void Process()
    {
        string layers = context.Request["LAYERS"];
        string styles = context.Request["STYLES"];
        string srs = context.Request["SRS"];
        BoundingBox bbox = ParseBBox(context.Request["BBOX"]);
        System.Drawing.Imaging.ImageCodecInfo imageEncoder = GetEncoderInfo(context.Request.Params["FORMAT"]);
        int width = int.Parse(context.Request["WIDTH"]);
        int height = int.Parse(context.Request["HEIGHT"]);

        string[] layerNames = layers.Split(',');
        if (layerNames.Length == 0)
            throw new InvalidDataException("Layers设置不正确");

        string[] styleNames = styles.Split(',');
        if (!string.IsNullOrEmpty(styles) && styleNames.Length != layerNames.Length)
            throw new InvalidDataException("Styles设置不正确");

        Map map = new Map(new System.Drawing.Size(width, height));
        BuildMapLayers(srs, layerNames, styleNames, map);

        map.PixelAspectRatio = ((double)width / (double)height) / (bbox.Width / bbox.Height);
        map.Center = bbox.GetCentroid();
        map.Zoom = bbox.Width;

        using (System.Drawing.Image img = map.GetMap())
        {
            //Png can't stream directy. Going through a memorystream instead
            System.IO.MemoryStream MS = new System.IO.MemoryStream();
            img.Save(MS, imageEncoder, null);
            img.Dispose();
            byte[] buffer = MS.ToArray();
            context.Response.ContentType = imageEncoder.MimeType;
            context.Response.OutputStream.Write(buffer, 0, buffer.Length);
        }
    }

    private void BuildMapLayers(string srs, string[] layerNames, string[] styleNames, Map map)
    {
        for (int i = 0; i < layerNames.Length; ++i)
        {
            string layerName = layerNames[i];
            VectorLayer layer = new VectorLayer(layerName);
            InitLayer(srs, layerName, layer);
            layer.Style = BuildStyle(layerName, i, styleNames);
            map.Layers.Add(layer);
        }
    }

    private void InitLayer(string srs, string layerName, VectorLayer layer)
    {
        string dsf = Config.GetDataProviderFile(layerName);
        XmlDocument doc = new XmlDocument();
        doc.Load(dsf);
        XmlNode n = doc.SelectSingleNode("/Data/EsriShape");
        IProvider pvd = null;
        ICoordinateTransformation trans = null;
        if (n != null)
        {
            ShapeFile sf = new ShapeFile(Config.GetEsriShapeFile(n.InnerText));
            if (sf.CoordinateSystem != null)
                trans = new CoordinateTransformationFactory()
                    .CreateFromCoordinateSystems(sf.CoordinateSystem, GetCoordinateSystemBySRS(srs));
            pvd = sf;
        }
        n = doc.SelectSingleNode("/Data/WFS");
        if (n != null)
        {
            string geo = n.SelectSingleNode("GeoProperty").InnerText;
            string srcSrs = n.SelectSingleNode("SRS").InnerText;
            pvd = new WfsProvider(n.SelectSingleNode("Uri").InnerText
                , n.SelectSingleNode("Name").InnerText
                , new string[] { geo }, geo, srcSrs, BuildBBox((XmlElement)n.SelectSingleNode("BoundingBox")));
            if (srcSrs != srs)
                trans = new CoordinateTransformationFactory()
                        .CreateFromCoordinateSystems(GetCoordinateSystemBySRS(srcSrs), GetCoordinateSystemBySRS(srs));
        }
        if (pvd == null)
            throw new NotImplementedException();
        layer.DataSource = pvd;
        layer.CoordinateTransformation = trans;
    }

    private BoundingBox BuildBBox(XmlElement xmlNode)
    {
        return new BoundingBox(double.Parse(xmlNode.GetAttribute("minx"))
            , double.Parse(xmlNode.GetAttribute("miny"))
            , double.Parse(xmlNode.GetAttribute("maxx"))
            , double.Parse(xmlNode.GetAttribute("maxy")));
    }

    private ICoordinateSystem GetCoordinateSystemBySRS(string srs)
    {
        string[] ss = srs.Split(':');
        if (ss.Length != 2)
            throw new InvalidDataException();

        foreach (string i in File.ReadAllLines(Config.GetSridFile()))
        {
            if (i.StartsWith(ss[1]))
                return new SharpMap.CoordinateSystems.CoordinateSystemFactory()
                    .CreateFromWkt(i.Substring(i.IndexOf(';') + 1));
        }

        throw new NotImplementedException("还不支持这种SRS" + srs);
    }

    private VectorStyle BuildStyle(string layerName, int i, string[] styleNames)
    {
        string styleName = (styleNames.Length > i ? styleNames[i] : "");
        if (string.IsNullOrEmpty(styleName))
            styleName = GetDefaultStyleName(layerName);

        XmlDocument doc = new XmlDocument();
        doc.Load(Config.GetSldFile(styleName));

        XmlNode n = doc.DocumentElement.SelectSingleNode("NamedLayer/UserStyle/FeatureTypeStyle/Rule");

        if (n == null)
            throw new Exception("数据错误");

        XmlNode sn = n.SelectSingleNode("LineSymbolizer");
        if (sn != null)
            return ParseLineSymbolStyle(sn);
        sn = n.SelectSingleNode("PointSymbolizer");
        if (sn != null)
            return ParsePointSymbolStyle(sn);
        sn = n.SelectSingleNode("PolygonSymbolizer");
        if (sn != null)
            return ParsePolygonSymbolStyle(sn);

        return null;
    }

    private VectorStyle ParseLineSymbolStyle(XmlNode sn)
    {
        VectorStyle ret = new VectorStyle();
        ret.Line = GetStroke(sn);
        return ret;
    }

    private static Pen GetStroke(XmlNode sn)
    {
        Color clr = Color.Black;
        XmlNode n = sn.SelectSingleNode("Stroke/CssParameter[@name='stroke']");
        if (n != null)
            clr = ParseColorFromCSS(n.InnerText);
        int width = 1;
        n = sn.SelectSingleNode("Stroke/CssParameter[@name='stroke-width']");
        if (n != null)
            width = int.Parse(n.InnerText);
        Pen stroke = new Pen(clr, width);
        return stroke;
    }

    private static Color ParseColorFromCSS(string css)
    {
        return Color.FromArgb(Convert.ToInt32(css.Substring(1, 2), 16)
            , Convert.ToInt32(css.Substring(3, 2), 16)
            , Convert.ToInt32(css.Substring(5, 2), 16));
    }

    private VectorStyle ParsePointSymbolStyle(XmlNode sn)
    {
        VectorStyle ret = new VectorStyle();
        float size = 6;
        float rotation = 0;

        XmlNode n = sn.SelectSingleNode("Graphic/Size");
        if (n != null)
            size = float.Parse(n.InnerText);

        n = sn.SelectSingleNode("Graphic/Rotation");
        if (n != null)
            rotation = float.Parse(n.InnerText);

        //Set WellKnownName Mark
        Bitmap bmp = null;
        n = sn.SelectSingleNode("Graphic/Mark/WellKnownName");
        if (n != null)
        {
            string wkn = n.InnerText;
            n = sn.SelectSingleNode("Graphic/Mark");
            bmp = GetWellKnownBitmap(wkn, size, ParseFill(n), GetStroke(n));
        }

        //or set ExternalGraphic
        if (bmp == null)
        {
            n = sn.SelectSingleNode("Graphic/ExternalGraphic/OnlineResource@xlink:href");
        }

        if (bmp != null)
            ret.Symbol = bmp;

        ret.SymbolOffset = new PointF(-size, -size);
        ret.SymbolRotation = rotation;
        return ret;
    }

    private Bitmap GetWellKnownBitmap(string wkn, float size, Brush brush, Pen pen)
    {
        GraphicsPath gp = GetWellKnownSymbol(wkn);

        if (gp == null)
            return null;

        using (gp)
        {
            Matrix m = new Matrix();
            m.Scale(size, size);
            gp.Transform(m);

            Bitmap ret = new Bitmap((int)size, (int)size);
            using (Graphics g = Graphics.FromImage(ret))
            {
                if (brush != null)
                    g.FillPath(brush, gp);
                if (pen != null)
                    g.DrawPath(pen, gp);
            }
            return ret;
        }
    }

    private GraphicsPath GetWellKnownSymbol(string wkn)
    {
        string wkf = Config.GetWktFile(wkn + ".txt");
        if (!File.Exists(wkf))
            return null;

        string[] ls = File.ReadAllLines(wkf);
        if (ls.Length < 1)
            throw new InvalidDataException();
        if (ls[0].Equals("Ellipse", StringComparison.OrdinalIgnoreCase))
            return BuildEllipsePath(ls);
        if (ls[0].Equals("Vertex", StringComparison.OrdinalIgnoreCase))
            return BuildVertexPath(ls);
        throw new NotImplementedException();
    }


    private GraphicsPath BuildVertexPath(string[] ls)
    {
        if (ls.Length < 5)
            throw new InvalidDataException();

        PointF[] vers = new PointF[ls.Length - 1];
        for (int i = 1; i < ls.Length; ++i)
        {
            vers[i - 1] = ParsePoint(ls[i]);
        }
        GraphicsPath ret = new GraphicsPath();
        ret.AddLines(vers);
        return ret;
    }

    private PointF ParsePoint(string s)
    {
        string[] ss = s.Split(',');
        if (ss.Length != 2)
            throw new InvalidDataException();
        return new PointF(float.Parse(ss[0]), float.Parse(ss[1]));
    }

    private GraphicsPath BuildEllipsePath(string[] ls)
    {
        GraphicsPath ret = new GraphicsPath();
        ret.AddEllipse(0, 0, 1, 1);
        return ret;
    }

    private VectorStyle ParsePolygonSymbolStyle(XmlNode sn)
    {
        VectorStyle ret = new VectorStyle();

        Brush b = ParseFill(sn);
        if (b != null)
            ret.Fill = b;

        return ret;
    }

    private static Brush ParseFill(XmlNode sn)
    {
        Brush b = null;
        XmlNode n = sn.SelectSingleNode("Fill/CssParameter[@name='fill']");
        if (n != null)
            b = new SolidBrush(ParseColorFromCSS(n.InnerText));
        return b;
    }

    private string GetDefaultStyleName(string layerName)
    {
        string lf = Config.GetWmsLayerFile(layerName);
        XmlDocument doc = new XmlDocument();
        doc.Load(lf);
        foreach (XmlNode i in doc.SelectNodes("Layer/Style"))
        {
            return i.InnerText;
        }
        throw new InvalidOperationException();
    }

    //部分代码抄袭了SharpMap
    private static System.Drawing.Imaging.ImageCodecInfo GetEncoderInfo(String mimeType)
    {
        foreach (System.Drawing.Imaging.ImageCodecInfo encoder in System.Drawing.Imaging.ImageCodecInfo.GetImageEncoders())
            if (encoder.MimeType == mimeType)
                return encoder;
        return null;
    }

    private BoundingBox ParseBBox(string p)
    {
        string[] ss = p.Split(',');
        return new BoundingBox(double.Parse(ss[0]), double.Parse(ss[1]), double.Parse(ss[2]), double.Parse(ss[3]));
    }
}
