﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Ogc.Wms;
using System.Net;
using System.IO;

namespace WmsBrowser
{
    public partial class MainForm : Form
    {
        private string ServerUri
        {
            get { return serverUrl.Text.Trim(); }
        }

        private string GetMapParameters
        {
            set { getMapUrl.Text = value; }
            get { return getMapUrl.Text; }
        }

        public MainForm()
        {
            InitializeComponent();
            UpdateCtrlsEnable();
        }

        private void getCapabilities_Click(object sender, EventArgs e)
        {
            try
            {
                Ogc.Wms.Capabilities si = new Ogc.Wms.Client(ServerUri).GetCapabilities();
                this.Text = "Wms Server " + si + " is Opened.";
                UpdateFormat(si.Capability.MapFormats);
                Layer l = si.Capability.Layer;
                UpdateSRS(l.Srss);
                Box bbox = l.LatLonBoundingBox;
                SetBBox(bbox);
                layers.Items.Clear();
                layerTree.Nodes.Clear();
                UpdateLayerTree(l, layerTree.Nodes);
            }
            catch (WebException ex)
            {
                if (ex.Response != null)
                    new HtmlForm(ex.Response.GetResponseStream()).ShowDialog();
                else
                    MessageBox.Show(ex.Message);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void UpdateSRS(string[] srss)
        {
            srs.Items.Clear();
            foreach (string i in srss)
            {
                srs.Items.Add(i);
            }

            if (srs.Items.Count == 0)
                srs.Items.Add("EPSG:4326");

            srs.SelectedItem = "EPSG:4326";

            if (srs.SelectedItem == null)
                srs.SelectedIndex = 0;
        }

        private void UpdateFormat(string[] fmts)
        {
            format.Items.Clear();
            foreach (string i in fmts)
            {
                format.Items.Add(i);
            }

            if (format.Items.Count > 0)
                format.SelectedIndex = 0;
        }

        private void UpdateLayerTree(Ogc.Wms.Layer layer, TreeNodeCollection nodes)
        {
            foreach (Layer i in layer.Layers)
            {
                TreeNode tn = new TreeNode(i.ToString());
                tn.Tag = i;
                UpdateLayerTree(i, tn.Nodes);
                nodes.Add(tn);
            }
        }

        private void getMap_Click(object sender, EventArgs e)
        {
            try
            {
                new MapForm(new Ogc.Wms.Client(ServerUri).GetMap(GetMapParameters)).ShowDialog();
            }
            catch (WebException ex)
            {
                if (ex.Response != null)
                    new HtmlForm(ex.Response.GetResponseStream()).ShowDialog();
                else
                    MessageBox.Show(ex.Message);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void builParameters_Click(object sender, EventArgs e)
        {
            RequestGetMapParameters builder = new RequestGetMapParameters();

            builder.MapSize = new Size((int)mapWidth.Value, (int)mapHeight.Value);
            builder.BBox = new double[] {
                (double)minX.Value,(double)minY.Value,(double)maxX.Value,(double)maxY.Value
            };
            builder.SRS = srs.Text;
            builder.Format = format.Text;
            foreach (LayerHolder i in layers.Items)
            {
                if (!string.IsNullOrEmpty(i.StyleName))
                    builder.AddLayer(i.LayerName, i.StyleName);
                else
                    builder.AddLayer(i.Layer);
            }
            GetMapParameters = builder.ToString();
        }

        private void toolTip1_Popup(object sender, PopupEventArgs e)
        {

        }

        private void layerTree_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            AddSelectedLayer();
        }

        private void AddSelectedLayer()
        {
            if(layerTree.SelectedNode == null)
                return;

            Layer l = layerTree.SelectedNode.Tag as Layer;

            if (l == null)
                return;

            layers.Items.Add(new LayerHolder(l,null));

            layers.SelectedIndex = layers.Items.Count - 1;
        }

        private void layers_SelectedIndexChanged(object sender, EventArgs e)
        {
            LayerHolder l = layers.SelectedItem as LayerHolder;
            if (l == null)
                return;

            UpdateStyles(l.Layer.Styles);
        }

        private void UpdateStyles(Style[] styles)
        {
            style.Items.Clear();

            foreach (Style i in styles)
            {
                style.Items.Add(i);
            }

            if (style.Items.Count > 0)
                style.SelectedIndex = 0;
        }

        private void layers_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            RemoveSelectedLayer();
        }

        private void RemoveSelectedLayer()
        {
            int i = layers.SelectedIndex;
            if (i < 0)
                return;

            layers.Items.RemoveAt(i);

            if (i < layers.Items.Count)
                layers.SelectedIndex = i;
            else
                style.Items.Clear();
        }

        private void style_SelectedIndexChanged(object sender, EventArgs e)
        {
            LayerHolder l = layers.SelectedItem as LayerHolder;
            if (l == null)
                return;

            Style s = style.SelectedItem as Style;
            if (s == null)
                return;
            l.Style = s;
        }

        private void addLayer_Click(object sender, EventArgs e)
        {
            AddSelectedLayer();
        }

        private void removeLayer_Click(object sender, EventArgs e)
        {
            RemoveSelectedLayer();
        }

        private void moveUp_Click(object sender, EventArgs e)
        {
            int i = layers.SelectedIndex;
            if (i <= 0)
                return;

            object o = layers.SelectedItem;
            RemoveSelectedLayer();
            --i;
            layers.Items.Insert(i, o);
            layers.SelectedIndex = i;
        }

        private void moveDown_Click(object sender, EventArgs e)
        {
            int i = layers.SelectedIndex;
            if (i >= layers.Items.Count - 1)
                return;

            object o = layers.SelectedItem;
            RemoveSelectedLayer();
            ++i;
            layers.Items.Insert(i, o);
            layers.SelectedIndex = i;
        }

        private void useBBox_Click(object sender, EventArgs e)
        {
            LayerHolder l = layers.SelectedItem as LayerHolder;
            if (l == null)
                return;
            UpdateSRS(l.Layer.Srss);
            if (l.Layer.BoundingBoxs.Length > 0 && !SetBoundingBox(l.Layer.BoundingBoxs[0]))
                SetBBox(l.Layer.LatLonBoundingBox);
        }

        private bool SetBoundingBox(BoundingBox bbox)
        {
            if (!bbox.Available)
                return false;
            SetBBox(bbox);
            srs.SelectedItem = bbox.SRS;
            return true;
        }

        private bool SetBBox(Box bbox)
        {
            if (!bbox.Available)
                return false;
            minX.Value = (decimal)bbox.MinX;
            minY.Value = (decimal)bbox.MinY;
            maxX.Value = (decimal)bbox.MaxX;
            maxY.Value = (decimal)bbox.MaxY;
            return true;
        }

        private void getMapUrl_TextChanged(object sender, EventArgs e)
        {
            UpdateCtrlsEnable();
        }

        private void UpdateCtrlsEnable()
        {
            getMap.Enabled = !string.IsNullOrEmpty(GetMapParameters);
            getCapabilities.Enabled = !string.IsNullOrEmpty(ServerUri);
        }

        private void serverUrl_TextChanged(object sender, EventArgs e)
        {
            UpdateCtrlsEnable();
        }
    }

    class LayerHolder
    {
        Layer layer;
        public Layer Layer
        {
            get { return layer; }
            set { layer = value; }
        }

        Style style;
        public Style Style
        {
            get { return style; }
            set { style = value; }
        }

        public string LayerName
        {
            get
            {
                if (layer != null)
                    return layer.Name;
                return "";
            }
        }

        public string StyleName
        {
            get
            {
                if (style != null)
                    return style.Name;
                return "";
            }
        }

        public LayerHolder(Layer layer, Style style)
        {
            this.layer = layer;
            this.style = style;
        }

        public override string ToString()
        {
            return layer.ToString();
        }
    }
}